/*
  A simple 2D hydro code
  (C) Romain Teyssier : CEA/IRFU           -- original F90 code
  (C) Pierre-Francois Lavallee : IDRIS      -- original F90 code
  (C) Guillaume Colin de Verdiere : CEA/DAM -- for the C version
  (C) Dimitri Lecas : IDRIS                 -- Pthreads version
*/
#include <stdio.h>
#include <time.h>

#include "parametres.h"
#include "hydro_funcs.h"
#include "vtkfile.h"
#include "compute_deltat.h"
#include "hydro_godunov.h"
#include "utils.h"

#include "pthr.h"
#include "pthrsingle.h"

/* private data */
typedef struct _private {
  hydrovarwork_t Hvw;
  hydrowork_t Hw;
} pthr_mainprivate_t;

/* data shared between main & mainthread function */
typedef struct _shared {
  double dt;
  long nvtk;
  long time_output;
  double start_iter;
  double end_iter;
  double next_output_time;
  long flag;
  char outnum[80];
} pthr_mainshared_t;

hydroparam_t H;
hydrovar_t Hv;                  // nvar
//hydrovarwork_t Hvw;             // nvar
//hydrowork_t Hw;
unsigned long flops = 0;
pthr_mainshared_t shared;

/* Main function for threads */
void * mainthread(void * ptr)
{
  int status;
  pthr_private_t * privateData;
  pthr_mainprivate_t workData;
#pragma warning (disable:810)
  int me = (int) ptr;
#pragma warning (default:810)

  /* Init private part */
  privateData = (pthr_private_t *) malloc(sizeof(pthr_private_t));
  CheckMalloc(privateData);
  /* Associate with this to retrieve later my rank (me) */
  status = pthread_setspecific(pthr_shared.this, privateData);
  if (status != 0)
    PthrErrAbort(status, "pthread_setspecific");
  privateData->me = me;
  /*fprintf(stdout, "Me = %d\n", me);*/

  while ((H.t < H.tend) && (H.nstep < H.nstepmax))
    {
      if (pthr_single_wait( &(pthr_shared.singlewait) ) == -1 )
	{
	  /* Single wait section */
	  shared.start_iter = cclock();
	  shared.outnum[0] = 0;
	  flops = 0;
	  pthr_single_wait( &(pthr_shared.singlewait) );
	}
      if ((H.nstep % 2) == 0)
	{
	  compute_deltat(&(shared.dt), H, &(workData.Hw), &Hv, &(workData.Hvw));
	  if (pthr_single_wait( &(pthr_shared.singlewait) ) == -1)
	    {
	      /* Single wait section */
	      if (H.nstep == 0) {
		shared.dt = shared.dt / 2.0;
	      }
	      pthr_single_wait( &(pthr_shared.singlewait) );
	    }
	}

      if ((H.nstep % 2) == 0) {
	hydro_godunov(1, shared.dt, H, &Hv, &(workData.Hw), &(workData.Hvw));
	hydro_godunov(2, shared.dt, H, &Hv, &(workData.Hw), &(workData.Hvw));
      } else {
	hydro_godunov(2, shared.dt, H, &Hv, &(workData.Hw), &(workData.Hvw));
	hydro_godunov(1, shared.dt, H, &Hv, &(workData.Hw), &(workData.Hvw));
      }

      if (pthr_single_wait( &(pthr_shared.singlewait) ) == -1 )
	{
	  /* Single wait section */
	  shared.end_iter = cclock();
	  H.nstep++;
	  H.t += shared.dt;

	  if (flops > 0) {
	    double iter_time = (double) (shared.end_iter - shared.start_iter);
	    if (iter_time > 1.e-9) {
	      double mflops = (double) flops / (double) 1.e+6 / iter_time;
	      sprintf(shared.outnum, "%s {%.3f Mflops} (%.3fs)", shared.outnum, mflops, iter_time);
	    }
	  } else {
	    double iter_time = (double) (shared.end_iter - shared.start_iter);
	    sprintf(shared.outnum, "%s (%.3fs)", shared.outnum, iter_time);
	  }
	  if (shared.time_output == 0) {
	    if ((H.nstep % H.noutput) == 0) {
	      vtkfile(++(shared.nvtk), H, &Hv);
	      sprintf(shared.outnum, "%s [%04ld]", shared.outnum, shared.nvtk);
	    }
	  } else {
	    if (H.t >= shared.next_output_time) {
	      vtkfile(++(shared.nvtk), H, &Hv);
	      shared.next_output_time = shared.next_output_time + H.dtoutput;
	      sprintf(shared.outnum, "%s [%04ld]", shared.outnum, shared.nvtk);
	    }
	  }
	  fprintf(stdout, "--> step=%-4ld %12.5e, %10.5e %s\n", H.nstep, H.t, shared.dt, shared.outnum);
	  pthr_single_wait( &(pthr_shared.singlewait) );
	}
    }   // end while loop
  free(privateData);
  return NULL;
}


int
main(int argc, char **argv)
{
  int iter,status;
  pthread_t * threadid;

  // double output_time = 0.0;
  double start_time = 0, end_time = 0;
  double elaps = 0;

  start_time = cclock();
  process_args(argc, argv, &H);
  hydro_init(&H, &Hv);
  PRINTUOLD(H, &Hv);

  /* Init shared value between mains */
  shared.dt = 0;
  shared.nvtk = 0;
  shared.time_output = 0;
  shared.start_iter = 0;
  shared.end_iter = 0;
  shared.next_output_time = 0;
  shared.flag = 0;

  /* Init global shared */
  pthr_init(H.nbthreads);

  printf("Hydro starts on %ld thread(s)\n",H.nbthreads);

  // vtkfile(shared.nvtk, H, &Hv);
  if (H.dtoutput > 0)
    {
      // outputs are in physical time not in time steps
      shared.time_output = 1;
      shared.next_output_time = shared.next_output_time + H.dtoutput;
    }

  /* launch son threads */
  if (H.nbthreads > 1) {
    threadid = (pthread_t *) malloc( (H.nbthreads-1)*sizeof(pthread_t) );
    CheckMalloc(threadid);}
  for (iter=0; iter<H.nbthreads-1; iter++)
    {
      status = pthread_create( &(threadid[iter]),
			       NULL,
			       mainthread,
			       (void *) iter);
      if (status != 0)
	PthrErrAbort(status,"pthread_create");
    }
  /* I'am also a thread */
  iter = H.nbthreads-1;
  mainthread((void *) iter);

  /* Wait for the end of all threads */
  for (iter=0; iter<H.nbthreads-1;iter++)
    {
      status = pthread_join(threadid[iter],NULL);
      if (status != 0)
	PthrErrAbort(status,"pthread_join");
    }

  hydro_finish(H, &Hv);
  end_time = cclock();
  elaps = (double) (end_time - start_time);
  timeToString(shared.outnum, elaps);

  fprintf(stdout, "Hydro ends in %ss (%.3lf) with %ld threads.\n", shared.outnum, elaps, H.nbthreads);
  if (H.nbthreads > 1) {
    free(threadid); }
  pthr_destroy();
  return 0;
}
