!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! -*- Mode: F90 -*- !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! module_hydro_principal.f90 --- 
!!!!
!! subroutine init_hydro
!! subroutine cmpdt(dt)
!! subroutine godunov(idim,dt)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module hydro_principal

contains

subroutine init_hydro
  !$ use OMP_LIB
  use hydro_commons
  use hydro_const
  use hydro_parameters
  use hydro_utils
  use omp_parameters
  use omp_domain_decomposition
  implicit none

  ! Local variables
  integer(kind=prec_int) :: i,j
  integer :: distribxy

  !$OMP SINGLE
  imin=1
  imax=nx+4
  jmin=1
  jmax=ny+4
  allocate(uold(imin:imax,jmin:jmax,1:nvar))
  nb_th =omp_get_num_threads()
  !$OMP END SINGLE

  ! computation of local sub-domains of dimension
  ! X : iminloc1->imaxloc1
  ! Y : jminloc2->jmaxloc2
  ! distribxy = 1 : decomposition suivant X
  distribxy=1
  call bornes_locales(distribxy,iminloc1,imaxloc1,jminloc1,jmaxloc1)
  ! distribxy = 2 : decomposition suivant Y
  distribxy=2
  call bornes_locales(distribxy,iminloc2,imaxloc2,jminloc2,jmaxloc2)
  call decomp_y

  ! Initial conditions in grid interior
  ! Warning: conservative variables U = (rho, rhou, rhov, E)

!!$  ! Jet
!!$  do j=jmin+2,jmax-2
!!$     do i=imin+2,imax-2
!!$        uold(i,j,ID)=1.0
!!$        uold(i,j,IU)=0.0
!!$        uold(i,j,IV)=0.0
!!$        uold(i,j,IP)=1.0/(gamma-1.0)
!!$     end do
!!$  end do

  ! Wind tunnel with point explosion
  do j=jminloc,jmaxloc
     do i=iminloc,imaxloc
        uold(i,j,ID)=1.0
        uold(i,j,IU)=0.0
        uold(i,j,IV)=0.0
        uold(i,j,IP)=1.d-5
     end do
  end do

  !$OMP BARRIER
  !$OMP SINGLE
  uold(imin+2,jmin+2,IP)=1./dx/dx
  !$OMP END SINGLE

!!$  ! 1D Sod test
!!$  do j=jmin+2,jmax-2
!!$     do i=imin+2,imax/2
!!$        uold(i,j,ID)=1.0
!!$        uold(i,j,IU)=0.0
!!$        uold(i,j,IV)=0.0
!!$        uold(i,j,IP)=1.0/(gamma-1.0)
!!$     end do
!!$     do i=imax/2+1,imax-2
!!$        uold(i,j,ID)=0.125
!!$        uold(i,j,IU)=0.0
!!$        uold(i,j,IV)=0.0
!!$        uold(i,j,IP)=0.1/(gamma-1.0)
!!$     end do
!!$  end do
end subroutine init_hydro


subroutine cmpdt(dt)
  use hydro_commons
  use hydro_const
  use hydro_parameters
  use hydro_utils
  implicit none

  ! Dummy arguments
  real(kind=prec_real), intent(out) :: dt  
  ! Local variables
  integer(kind=prec_int) :: i,j
  real(kind=prec_real)   :: cournox=0,cournoy=0,eken
  real(kind=prec_real)   :: cournox_loc, cournoy_loc
  real(kind=prec_real),  dimension(:,:), allocatable   :: q
  real(kind=prec_real),  dimension(:)  , allocatable   :: e,c

  ! compute time step on grid interior
  cournox_loc = zero
  cournoy_loc = zero
  !$OMP SINGLE
  cournox = zero
  cournoy = zero
  !$OMP END SINGLE

  call decomp_y

  allocate(q(iminloc:imaxloc,1:IP))
  allocate(e(iminloc:imaxloc),c(iminloc:imaxloc))

  do j=jminloc,jmaxloc
     do i=iminloc,imaxloc
        q(i,ID) = max(uold(i,j,ID),smallr)
        q(i,IU) = uold(i,j,IU)/q(i,ID)
        q(i,IV) = uold(i,j,IV)/q(i,ID)
        eken = half*(q(i,IU)**2+q(i,IV)**2)
        q(i,IP) = uold(i,j,IP)/q(i,ID) - eken
        e(i)=q(i,IP)
     end do

     call eos(q,e,c)
  
     cournox_loc=max(cournox_loc,maxval(c(iminloc:imaxloc)+abs(q(iminloc:imaxloc,IU))))
     cournoy_loc=max(cournoy_loc,maxval(c(iminloc:imaxloc)+abs(q(iminloc:imaxloc,IV))))

  end do

  !$OMP ATOMIC
  cournox=max(cournox,cournox_loc)
  !$OMP ATOMIC
  cournoy=max(cournoy,cournoy_loc)

  deallocate(q,e,c)
  !$OMP BARRIER  

  !$OMP SINGLE
  dt = courant_factor*dx/max(cournox,cournoy,smallc)
  !$OMP END SINGLE
end subroutine cmpdt


subroutine godunov(idim,dt)
  use hydro_commons
  use hydro_const
  use hydro_parameters
  use hydro_utils
  use hydro_work_space
  use omp_parameters
  implicit none

  ! Dummy arguments
  integer(kind=prec_int), intent(in) :: idim
  real(kind=prec_real),   intent(in) :: dt
  ! Local variables
  integer(kind=prec_int) :: i,j,in
  real(kind=prec_real)   :: dtdx
  real(kind=prec_real)   :: tps_elapsed
  integer(kind=prec_int) :: nbp_init, nbp_final, nbp_max, freq_p

  ! constant
  dtdx=dt/dx

  ! Update boundary conditions
  call make_boundary(idim)

  if (idim==1)then
     ! Decomposition suivant Y
     call decomp_y

     ! Allocate work space for 1D sweeps
     call allocate_work_space(iminloc-2,imaxloc+2)

     do j=jminloc,jmaxloc
        ! Gather conservative variables
        do i=iminloc-2,imaxloc+2
           u(i,ID)=uold(i,j,ID)
           u(i,IU)=uold(i,j,IU)
           u(i,IV)=uold(i,j,IV)
           u(i,IP)=uold(i,j,IP)
        end do
        if(nvar>4)then
           do in = 5,nvar
              do i=iminloc-2,imaxloc+2
                 u(i,in)=uold(i,j,in)
              end do
           end do
        end if

        ! Convert to primitive variables
        call constoprim(u,q,c)

        ! Characteristic tracing
        call trace(q,dq,c,qxm,qxp,dtdx)

        do in = 1,nvar
           do i=iminloc-2,imaxloc-1
              qleft (i,in)=qxm(i+1,in)
              qright(i,in)=qxp(i+2,in)
           end do
        end do

        ! Solve Riemann problem at interfaces
        call riemann(qleft,qright,qgdnv, &
                     rl,ul,pl,cl,wl,rr,ur,pr,cr,wr,ro,uo,po,co,wo, &
                     rstar,ustar,pstar,cstar,sgnm,spin,spout, &
                     ushock,frac,scr,delp,pold,ind,ind2)

        ! Compute fluxes
        call cmpflx(qgdnv,flux)

        do i=iminloc,imaxloc
           uold(i,j,ID)=u(i,ID)+(flux(i-2,ID)-flux(i-1,ID))*dtdx
           uold(i,j,IU)=u(i,IU)+(flux(i-2,IU)-flux(i-1,IU))*dtdx
           uold(i,j,IV)=u(i,IV)+(flux(i-2,IV)-flux(i-1,IV))*dtdx
           uold(i,j,IP)=u(i,IP)+(flux(i-2,IP)-flux(i-1,IP))*dtdx
        end do

        if(nvar>4)then
           do in = 5,nvar
              do i=iminloc,imaxloc
                 uold(i,j,in)=u(i,in)+(flux(i-2,in)-flux(i-1,in))*dtdx
              end do
           end do
        end if
     end do
     !$OMP BARRIER

     ! Deallocate work space
     call deallocate_work_space

  else
     ! Decomposition suivant X
     call decomp_x

     ! Allocate work space for 1D sweeps
     call allocate_work_space(jminloc-2,jmaxloc+2)

     do i=iminloc,imaxloc
        ! Gather conservative variables
        do j=jminloc-2,jmaxloc+2
           u(j,ID)=uold(i,j,ID)
           u(j,IU)=uold(i,j,IV)
           u(j,IV)=uold(i,j,IU)
           u(j,IP)=uold(i,j,IP)
        end do
        if(nvar>4)then
           do in = 5,nvar
              do j=jminloc-2,jmaxloc+2
                 u(j,in)=uold(i,j,in)
              end do
           end do
        end if

        ! Convert to primitive variables
        call constoprim(u,q,c)

        ! Characteristic tracing
        call trace(q,dq,c,qxm,qxp,dtdx)

        do in = 1, nvar
           do j = jminloc-2, jmaxloc-1
              qleft (j,in)=qxm(j+1,in)
              qright(j,in)=qxp(j+2,in)
           end do
        end do

        ! Solve Riemann problem at interfaces
        call riemann(qleft,qright,qgdnv, &
                     rl,ul,pl,cl,wl,rr,ur,pr,cr,wr,ro,uo,po,co,wo, &
                     rstar,ustar,pstar,cstar,sgnm,spin,spout, &
                     ushock,frac,scr,delp,pold,ind,ind2)

        ! Compute fluxes
        call cmpflx(qgdnv,flux)

        ! Update conservative variables 
        do j=jminloc,jmaxloc
           uold(i,j,ID)=u(j,ID)+(flux(j-2,ID)-flux(j-1,ID))*dtdx
           uold(i,j,IU)=u(j,IV)+(flux(j-2,IV)-flux(j-1,IV))*dtdx
           uold(i,j,IV)=u(j,IU)+(flux(j-2,IU)-flux(j-1,IU))*dtdx
           uold(i,j,IP)=u(j,IP)+(flux(j-2,IP)-flux(j-1,IP))*dtdx
        end do

        if(nvar>4)then
           do in = 5,nvar
              do j=jminloc,jmaxloc
                 uold(i,j,in)=u(j,in)+(flux(j-2,in)-flux(j-1,in))*dtdx
              end do
           end do
        end if
     end do

     ! Deallocate work space
     call deallocate_work_space

  end if

  !$OMP BARRIER

contains

  subroutine allocate_work_space(ii1,ii2)
    implicit none

    ! Dummy arguments
    integer(kind=prec_int), intent(in) :: ii1,ii2

    allocate(u  (ii1:ii2,1:nvar))
    allocate(q  (ii1:ii2,1:nvar))
    allocate(dq (ii1:ii2,1:nvar))
    allocate(qxm(ii1:ii2,1:nvar))
    allocate(qxp(ii1:ii2,1:nvar))
    allocate(c  (ii1:ii2))
    allocate(qleft (ii1:ii2-3,1:nvar))
    allocate(qright(ii1:ii2-3,1:nvar))
    allocate(qgdnv (ii1:ii2-3,1:nvar))
    allocate(flux  (ii1:ii2-3,1:nvar))
    allocate(rl    (ii1:ii2-3), ul   (ii1:ii2-3), pl   (ii1:ii2-3), cl    (ii1:ii2-3))
    allocate(rr    (ii1:ii2-3), ur   (ii1:ii2-3), pr   (ii1:ii2-3), cr    (ii1:ii2-3))
    allocate(ro    (ii1:ii2-3), uo   (ii1:ii2-3), po   (ii1:ii2-3), co    (ii1:ii2-3))
    allocate(rstar (ii1:ii2-3), ustar(ii1:ii2-3), pstar(ii1:ii2-3), cstar (ii1:ii2-3))
    allocate(wl    (ii1:ii2-3), wr   (ii1:ii2-3), wo   (ii1:ii2-3))
    allocate(sgnm  (ii1:ii2-3), spin (ii1:ii2-3), spout(ii1:ii2-3), ushock(ii1:ii2-3))
    allocate(frac  (ii1:ii2-3), scr  (ii1:ii2-3), delp (ii1:ii2-3), pold  (ii1:ii2-3))
    allocate(ind   (ii1:ii2-3), ind2 (ii1:ii2-3))
  end subroutine allocate_work_space

  subroutine deallocate_work_space
    deallocate(u,q,dq,qxm,qxp,c,qleft,qright,qgdnv,flux)
    deallocate(rl,ul,pl,cl)
    deallocate(rr,ur,pr,cr)  
    deallocate(ro,uo,po,co)  
    deallocate(rstar,ustar,pstar,cstar)
    deallocate(wl,wr,wo)
    deallocate(sgnm,spin,spout,ushock)
    deallocate(frac,scr,delp,pold)
    deallocate(ind,ind2)    
  end subroutine deallocate_work_space

end subroutine godunov

end module hydro_principal
