module omp_parameters
  integer :: num_th, nb_th
  integer :: nb_dims
  integer, dimension(:), allocatable :: dims, tab_sync 
  integer, dimension(:,:), allocatable :: tab_voisin
  !$OMP THREADPRIVATE(num_th)
end module omp_parameters

module omp_domain_decomposition
  use omp_parameters

  implicit none

contains
  
! Calcule et stocke les nb_premier nombres premiers inferieurs ou egal
! au nombre num passe en argument dans le tableau tab_premier.
! tab_premier(1:nb_premier)=[p_1=2,p_2=3,...,p_i,p_i+1,...] 
subroutine nombre_premier(num, nb_premier, tab_premier)
  integer, intent(in) :: num
  integer, intent(out) :: nb_premier
  integer, pointer, dimension(:), intent(out) :: tab_premier
  integer :: n, taille, i, j

  taille = num/2+1
  allocate(tab_premier(taille))
  i=1
  tab_premier(i)=2
  
  do n=3,num,2
     j = 2
     do while (j<=i)
        if (mod(n,tab_premier(j))==0) exit
        j = j + 1
     enddo
     if (j==i+1) then
        i=i+1
        tab_premier(i)=n
     endif
  enddo
  nb_premier=i
end subroutine nombre_premier

! Calcule la factorisation du nombre num en facteurs premiers
! Les coefficents sont stockes dans le tableau tab_fact
! Si num = a1*p_1+a4*p_4 alors tab_fact=[a1,0,0,a4,...]
subroutine factorisation(num, nb_premier, tab_premier, tab_fact)
  integer, intent(in) :: nb_premier, num
  integer, dimension(:), intent(in) :: tab_premier
  integer, pointer, dimension(:), intent(out) :: tab_fact
  integer :: i, num_loc
  
  num_loc=num
  allocate(tab_fact(nb_premier))
  tab_fact(1:nb_premier)=0
  
  do i=nb_premier,1,-1
     do while (mod(num_loc,tab_premier(i))==0)
        tab_fact(i)=tab_fact(i)+1
        num_loc=num_loc/tab_premier(i)
     enddo
  enddo
  
end subroutine factorisation

! Repartit de facon equilibree, la factorisation en nombre premier suivant
! les freedims dimensions de libre
subroutine repartition(freedims,nb_premier,tab_premier,tab_fact,dims_loc)
  integer, intent(in) :: freedims, nb_premier
  integer, dimension(:), intent(in) :: tab_premier, tab_fact
  integer, dimension(:), intent(out) :: dims_loc
  integer :: i, j, k
  
  dims_loc(1:freedims)=1
  ! Boucle sur le nombre de nombre premier
  do i=nb_premier,1,-1
     ! Boucle sur le facteur de multiplicite du ieme nombre premier
     do j=tab_fact(i),1,-1
        ! Dimension ayant la plus petite contribution
        k = minloc(dims_loc(1:freedims),1)
        dims_loc(k)=dims_loc(k)*tab_premier(i)
     enddo
  enddo
  
  ! ! ordre decroissant suivant les dimensions 
  ! do i=1,freedims
  !    do j=i,freedims
  !       if (dims_loc(i)<dims_loc(j)) then
  !          k=dims_loc(i)
  !          dims_loc(i)=dims_loc(j)
  !          dims_loc(j)=k
  !       endif
  !    enddo
  ! enddo

  ! ordre croissant suivant les dimensions 
  do i=1,freedims
     do j=i,freedims
        if (dims_loc(i)>dims_loc(j)) then
           k=dims_loc(i)
           dims_loc(i)=dims_loc(j)
           dims_loc(j)=k
        endif
     enddo
  enddo
  
end subroutine repartition

! Calcule la repartition des num threads suivant chacune des dimensions,
! le nombre de threads dans une dimension donnee (i) pouvant etre imposee
! via un dims(i)/=0
subroutine decomp_domaine(num,ndims,dims)
  integer, intent(in) :: num, ndims
  integer, dimension(:), intent(inout) :: dims
  integer :: freenum, freedims, nb_premier
  integer, dimension(:), allocatable :: dims_loc
  integer, dimension(:), pointer :: tab_premier, tab_fact
  integer :: i, j

  ! Determination du nombre de dimension libre (freedims) et compatibilite 
  ! avec les contraintes imposees dans les directions fixees.
  freedims=0
  freenum=num
  do i=1,ndims
     if (dims(i)==0) then
        freedims=freedims+1
     else
        if (mod(num,dims(i))/=0) then
           print*,"Nombre de threads et dimension(s) imposee(s) incompatibles : arret du code"
           stop
        else
           freenum=freenum/dims(i)
        endif
     endif
  end do
  
  if (freedims==0) then
     if (freenum/=1) then
        print*,"Nombre de threads et dimension(s) imposee(s) incompatibles : arret du code"
        stop
     else
        return   
     endif
  endif

  if (freenum<1) then
     print*,"Erreur decomposition domaine"
     stop
  else
     if (freenum == 1) then
        where(dims(:)==0) dims(:)=1
        return
     endif
  endif

  ! Il reste freenum threads a repartir suivant freedim dimensions
  allocate(dims_loc(freedims))
  call nombre_premier(freenum,nb_premier,tab_premier)
  call factorisation(freenum,nb_premier,tab_premier,tab_fact)
  call repartition(freedims,nb_premier,tab_premier,tab_fact,dims_loc)
  
  ! Recopie de la repartition finale dans le tableau dims 
  ! dans chacune des dimensions
  j=1
  do i=1,ndims
     if(dims(i)==0) then 
        dims(i)=dims_loc(j)
        j=j+1
     endif
  enddo
  deallocate(dims_loc)
end subroutine decomp_domaine

! Calcul des bornes locales a chaque sous-domaine
! distribxy = 1 => on distribue les threads suivant X
! distribxy = 2 => on distribue les threads suivant Y
! distribxy = 0 => on distribue les threads equitablement suivant X et Y
subroutine bornes_locales(distribxy)
  use hydro_commons
  use hydro_parameters
  !$ use OMP_LIB

  implicit none
  integer, intent(in) :: distribxy
  integer, dimension(:), allocatable :: coord
  integer :: i

  !$OMP SINGLE
  ! Domaine 2D
  nb_dims = 2
  allocate(dims(nb_dims))
  nb_th = omp_get_num_threads()
  ! Par defaut, decomposition equitablement repartie en X et Y
  dims(1:nb_dims) = 0
  select case (distribxy)
  case(1) ! Decomposition uniquement suivant X
     dims(1) = nb_th
  case(2) ! Decomposition uniquement suivant Y
     dims(2) = nb_th
  end select
  allocate(tab_sync(0:nb_th-1))
  allocate(tab_voisin(0:nb_th-1,4))
  !$OMP END SINGLE

  num_th = omp_get_thread_num()  
  allocate(coord(nb_dims))

  !$OMP SINGLE
  call decomp_domaine(nb_th,nb_dims,dims)
  !$OMP END SINGLE

  ! First dimension  : X -> 1..nx+4
  ! Second dimension : Y -> 1..ny+4
  ! Distribution des threads suivant X
  ! coord(1) = indice de 1er dim du sous-domaine
  ! coord(2) = indice de 2eme dim du sous-domaine
  select case (nb_dims)
  case(1)
     coord(1) = mod(num_th,dims(1))+1
  case(2)
     coord(1) = mod(num_th,dims(1))+1
     coord(2) = num_th/dims(1)+1
  case(3)
     coord(1) = mod(num_th,dims(1))+1
     coord(2) = mod(num_th/dims(1),dims(2))+1
     coord(3) = num_th/(dims(1)*dims(2))+1
  end select

  ! Mise a jour tableau tab_voisin
  ! avec -1 si pas de voisin
  ! avec le numero du thread voisin sinon
  ! left : tab_voisin(.,1)
  if (coord(1)==1) then
     tab_voisin(num_th,1) = -1
  else
      tab_voisin(num_th,1) = num_th-1
  endif
  ! right : tab_voisin(.,2)
  if (coord(1)==dims(1)) then
     tab_voisin(num_th,2) = -1    
  else
     tab_voisin(num_th,2) = num_th+1
  endif
  ! up : tab_voisin(.,3)
  if (coord(2)==dims(2)) then
     tab_voisin(num_th,3) = -1
  else
     tab_voisin(num_th,3) = num_th+dims(1)
  endif
  ! down : tab_voisin(.,4)
  if (coord(2)==1) then
     tab_voisin(num_th,4) = -1
  else
     tab_voisin(num_th,4) = num_th-dims(1)
  endif
  !$OMP BARRIER

  ! Update of local sub-domain boundary condition
  boundary_left_loc  = boundary_left
  boundary_right_loc = boundary_right
  boundary_up_loc    = boundary_up
  boundary_down_loc  = boundary_down
  if (coord(2)/=1)       boundary_down_loc  = 0
  if (coord(2)/=dims(2)) boundary_up_loc    = 0
  if (coord(1)/=1)       boundary_left_loc  = 0
  if (coord(1)/=dims(1)) boundary_right_loc = 0
  
  ! local sub-domain for each thread
  iminloc = ((coord(1)-1)*nx)/dims(1)+3 
  imaxloc = (coord(1)*nx)/dims(1)+2
  jminloc = ((coord(2)-1)*ny)/dims(2)+3 
  jmaxloc = (coord(2)*ny)/dims(2)+2
  
  ! do i=0,nb_th-1
  !    if (i==num_th) then
  !       print *,"Thread num : ",num_th
  !       print *,"iminloc, imaxloc = ", iminloc, imaxloc
  !       print *,"jminloc, jmaxloc = ", jminloc, jmaxloc
  !    endif
  !    !$OMP BARRIER
  ! enddo

  !$OMP BARRIER
  deallocate(coord)

end subroutine bornes_locales

subroutine sync_x
  use hydro_commons
  use hydro_parameters
  !$ use OMP_LIB

syncx :   do
           !$OMP FLUSH(tab_sync)
           ! no left 
           if (tab_voisin(num_th,1)==-1) then
              ! no right 
              if (tab_voisin(num_th,2)==-1) then
                 exit syncx
              else ! right OK ?
                 if (tab_sync(num_th)<=tab_sync(tab_voisin(num_th,2))) then 
                    exit syncx
                 endif
              endif
           else
              ! left OK
              if (tab_sync(num_th)<=tab_sync(tab_voisin(num_th,1))) then
                 ! no right ?
                 if (tab_voisin(num_th,2)==-1) then
                    exit syncx
                 else
                    ! right OK
                    if (tab_sync(num_th)<=tab_sync(tab_voisin(num_th,2))) then
                       exit syncx
                    endif
                 endif
              endif
           endif
        enddo syncx

end subroutine sync_x

subroutine sync_y
  use hydro_commons
  use hydro_parameters
  !$ use OMP_LIB
  
syncy :   do
     !$OMP FLUSH(tab_sync)
     ! no up
     if (tab_voisin(num_th,3)==-1) then
        ! no down 
        if (tab_voisin(num_th,4)==-1) then
           exit syncy
        else ! down OK ?
           if (tab_sync(num_th)<=tab_sync(tab_voisin(num_th,4))) then 
              exit syncy
           endif
        endif
     else
        ! up OK
        if (tab_sync(num_th)<=tab_sync(tab_voisin(num_th,3))) then
           ! no down ?
           if (tab_voisin(num_th,4)==-1) then
              exit syncy
           else
              ! down OK ?
              if (tab_sync(num_th)<=tab_sync(tab_voisin(num_th,4))) then
                 exit syncy
              endif
           endif
        endif
     endif
  enddo syncy
end subroutine sync_y
      
end module omp_domain_decomposition
